package dev.pfaff.glass.pthread;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;

import static dev.pfaff.glass.internal.Util.ADDRESS_SIZED_INT;
import static java.lang.foreign.ValueLayout.ADDRESS;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class PThread {
	private static final MethodHandle MH_pthread_threadid_np;
	private static final MethodHandle MH_pthread_self;
	/**
	 * The pointer to the main thread's {@code pthread_t}.
	 */
	private static final MemorySegment MS_main_thread;

	static {
		var lookup = SymbolLookup.libraryLookup("/usr/lib/system/libsystem_pthread.dylib", Arena.ofAuto());
		var linker = Linker.nativeLinker();
		MH_pthread_threadid_np = linker.downcallHandle(lookup.findOrThrow("pthread_threadid_np"),
				FunctionDescriptor.of(JAVA_INT, ADDRESS, ADDRESS.withTargetLayout(ADDRESS_SIZED_INT)),
				Linker.Option.critical(true));
		MH_pthread_self = linker.downcallHandle(lookup.findOrThrow("pthread_self"),
				FunctionDescriptor.of(ADDRESS),
				Linker.Option.critical(false));
		MS_main_thread = lookup.find("_main_thread")
				.orElseGet(() -> lookup.findOrThrow("_main_thread_ptr").get(ADDRESS, 0));
	}

	public static MemorySegment pthread_self() {
		try {
			return (MemorySegment) MH_pthread_self.invokeExact();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @param pthread_t may be null, in which case it is treated as though it were the result of calling
	 *                  {@link #pthread_self()}.
	 */
	public static int pthread_threadid_np(MemorySegment pthread_t) {
		var slot = new long[1];
		try {
			// NULL == self
			return (int) MH_pthread_threadid_np.invokeExact(pthread_t, MemorySegment.ofArray(slot));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public static boolean isOnMainThread() {
		return pthread_self().equals(MS_main_thread);
	}
}
